@@ -136,27 +136,153 @@ class Particles(xo.dress(ParticlesData, rename={
             'scalar_vars': scalar_vars,
             'per_particle_vars': per_particle_vars}
 
-    def to_dict(self):
-        dct = super().to_dict()
-        dct['delta'] = self.delta
-        dct['psigma'] = self.psigma
-        dct['rvv'] = self.rvv
-        dct['rpp'] = self.rpp
+    def to_dict(self, copy_to_cpu=True):
+        if copy_to_cpu:
+            cpobj = self.copy(_context=xo.context_default)
+            return cpobj.to_dict(copy_to_cpu=False)
+        else:
+            dct = super().to_dict()
+            dct['delta'] = self.delta
+            dct['psigma'] = self.psigma
+            dct['rvv'] = self.rvv
+            dct['rpp'] = self.rpp
         return dct
 
+    def to_pandas(self):
+        import pandas as pd
+        return pd.DataFrame(self.to_dict())
+
+    @classmethod
+    def from_pandas(cls, df, _context=None, _buffer=None, _offset=None):
+        dct = df.to_dict(orient='list')
+        for tt, nn in scalar_vars + size_vars:
+            if nn in dct.keys() and not np.isscalar(dct[nn]):
+                dct[nn] = dct[nn][0]
+        return cls(**dct, _context=_context, _buffer=_buffer, _offset=_offset)
+
+    @classmethod
+    def merge(cls, lst, _context=None, _buffer=None, _offset=None):
+
+        # TODO For now the merge is performed on CPU for add contexts.
+        # Slow for objects on GPU (transferred to CPU for the merge).
+
+        # Move everything to cpu
+        cpu_lst = []
+        for pp in lst:
+            assert isinstance(pp, cls)
+            if isinstance(pp._buffer.context, xo.ContextCpu):
+                cpu_lst.append(pp)
+            else:
+                cpu_lst.append(pp.copy(_context=xo.context_default))
+
+        # Check that scalar variable are compatible
+        for tt, nn in scalar_vars:
+            vals = [getattr(pp, nn) for pp in cpu_lst]
+            assert np.allclose(vals,
+                                getattr(cpu_lst[0], nn), rtol=0, atol=1e-14)
+
+        # Make new particle on CPU
+        capacity = np.sum([pp._capacity for pp in cpu_lst])
+        new_part_cpu = cls(_capacity=capacity)
+
+        # Copy scalar vars from first particle
+        for tt, nn in scalar_vars:
+            setattr(new_part_cpu, nn, getattr(cpu_lst[0], nn))
+
+        # Copy per-particle vars
+        first = 0
+        max_id_curr = -1
+        for pp in cpu_lst:
+            for tt, nn in per_particle_vars:
+                if not(nn == 'particle_id' or nn == 'parent_id'):
+                    getattr(new_part_cpu, nn)[
+                            first:first+pp._capacity] = getattr(pp, nn)
+
+            # Handle particle_ids and parent_ids
+            mask = pp.particle_id >= 0
+            new_id = pp.particle_id.copy()
+            new_parent_id = pp.parent_particle_id.copy()
+            if np.min(new_id[mask]) <= max_id_curr:
+                new_id[mask] += (max_id_curr + 1)
+                new_parent_id[mask] += (max_id_curr + 1)
+            new_part_cpu.particle_id[first:first+len(new_id)] = new_id
+            new_part_cpu.parent_particle_id[
+                    first:first+len(new_id)] = new_parent_id
+
+            max_id_curr = np.max(new_id)
+            first += pp._capacity
+
+        # Reorganize
+        new_part_cpu.reorganize()
+
+        # Copy to appropriate context
+        if _context is None and _buffer is None:
+            # Use constext of first particle
+            if isinstance(lst[0]._buffer.context, xo.ContextCpu):
+                new_part_cpu._buffer.context = lst[0]._buffer.context
+                return new_part_cpu
+            else:
+                return new_part_cpu.copy(_context=lst[0]._buffer.context)
+        else:
+            return new_part_cpu.copy(_context=_context, _buffer=_buffer,
+                                     _offset=_offset)
+
+    def filter(self, mask):
+
+        if isinstance(self._buffer.context, xo.ContextCpu):
+            self_cpu = self
+        else:
+            self_cpu = self.copy(_context=xo.context_default)
+
+        # copy mask to cpu is needed
+        if isinstance(mask, self._buffer.context.nplike_array_type):
+            mask = self._buffer.context.nparray_from_context_array(mask)
+
+            # Pyopencl returns int8 instead of bool
+            if (isinstance(self._buffer.context, xo.ContextPyopencl) and
+                mask.dtype==np.int8):
+                assert np.all((mask>=0) & (mask<=1))
+                mask = mask > 0
+
+        # Make new particle on CPU
+        test_x = self_cpu.x[mask]
+        capacity = len(test_x)
+        new_part_cpu = self.__class__(_capacity=capacity)
+
+        # Copy scalar vars from first particle
+        for tt, nn in scalar_vars:
+            setattr(new_part_cpu, nn, getattr(self_cpu, nn))
+
+        # Copy per-particle vars
+        for tt, nn in per_particle_vars:
+            getattr(new_part_cpu, nn)[:] = getattr(self_cpu, nn)[mask]
+
+        # Reorganize
+        new_part_cpu.reorganize()
+
+        # Copy to original context 
+        target_ctx = self._buffer.context
+        if isinstance(target_ctx, xo.ContextCpu):
+            new_part_cpu._buffer.context = target_ctx
+            return new_part_cpu
+        else:
+            return new_part_cpu.copy(_context=target_ctx)
+
     def __init__(self, **kwargs):
 
+        input_kwargs = kwargs.copy()
+
         if '_xobject' in kwargs.keys():
             # Initialize xobject
             self.xoinitialize(**kwargs)
         else:
-
             if any([nn in kwargs.keys() for tt, nn in per_particle_vars]):
                 # Needed to generate consistent longitudinal variables
                 pyparticles = Pyparticles(**kwargs)
 
                 part_dict = _pyparticles_to_xpart_dict(pyparticles)
-                if '_capacity' in kwargs.keys():
+                if ('_capacity' in kwargs.keys() and
+                         kwargs['_capacity'] is not None):
                     assert kwargs['_capacity'] >= part_dict['_capacity']
                 else:
                     kwargs['_capacity'] = part_dict['_capacity']
@@ -204,6 +330,21 @@ class Particles(xo.dress(ParticlesData, rename={
         self._num_active_particles = -1 # To be filled in only on CPU
         self._num_lost_particles = -1 # To be filled in only on CPU
 
+        # Force values provided by user if compatible
+        for nn in part_energy_varnames():
+            vvv = self._buffer.context.nparray_from_context_array(getattr(self, nn))
+            if nn in input_kwargs.keys():
+                if np.allclose(vvv, input_kwargs[nn], rtol=0, atol=1e-13):
+                    if np.isscalar(input_kwargs[nn]):
+                        getattr(self, "_"+nn)[:] = input_kwargs[nn]
+                    else:
+                        getattr(self, "_"+nn)[:] = (
+                                context.nparray_to_context_array(
+                                    np.array(input_kwargs[nn])))
+                else:
+                    raise ValueError(
+                            'What?! This should have been intercepted before!')
+
         if isinstance(self._buffer.context, xo.ContextCpu):
             # Particles always need to be organized to run on CPU
             self.reorganize()
